#!/usr/bin/env ruby

#
# iPhone Note Extractor
#
# This script extracts all notes from an iPhone backup into individual
# HTML files, given the path of the backup folder.
#
# The script will create two things in the current directory:
# the "notes.db" file that was backed up (the SQLite database),
# and a folder called "Notes" containing the generated HTML files.
#
# By default, the generated HTML files will use Marker Felt to
# emulate the appearance of Notes on the iPhone.  For those who
# object to Marker Felt and would rather have the notes use the
# iPhone-standard Helvetica, a "gruber-friendly" option is available
# to accommodate this preference.
# 
# Author::   Marshall Elfstrand (mailto:marshall@vengefulcow.com)
# Version::  1.1
#

require "rubygems"

require "fileutils"
require "sqlite3"
require "osx/foundation"


#
# Extracts notes from an iPhone backup.
#
# The simplest use of this class is to call the +extract_all_notes+ method
# as follows:
#
#     NoteExtractor.new.extract_all_notes
#
# For more control, the individual extraction methods may be called by
# themselves.
#
class NoteExtractor
  
  # Query used to retrieve notes from the SQLite database.
  NOTES_QUERY = <<-END_QUERY
    SELECT Note.title, note_bodies.data
    FROM Note INNER JOIN note_bodies ON Note.ROWID = note_bodies.note_id
  END_QUERY

  # Template used to render the HTML files.
  HTML_TEMPLATE = <<-END_TEMPLATE
    <!DOCTYPE html>
    <html lang="en">
      <head>
        <meta charset="UTF-8">
        <style type="text/css">
          body {
            background: #432E28;
            font-family: %s;
            font-size: 12pt;
          }
          #note {
            background: #FEFDBC;
            border: 1pt solid #7B6F45;
            color: black;
            margin: 3em auto 3em auto;
            padding: 2em;
            width: 36em;
            -webkit-box-shadow: 0em 0.5em 1.5em black;
          }
        </style>
        <title>%s</title>
      </head>
      <body>
        <div id="note">
          %s
        </div>
      </body>
    </html>
  END_TEMPLATE

  
  #
  # Initializes note extractor with default values.
  #
  def initialize
    gruber_friendly = false
  end #def


  #
  # Returns +true+ if Marker Felt is disabled in the output.
  #
  def gruber_friendly?
    @gruber_friendly
  end #def
  
  #
  # Set to +true+ to disable Marker Felt in the output.
  #
  def gruber_friendly=(value)
    @gruber_friendly = (value == true)
    if @gruber_friendly
      @font_family = 'Helvetica, sans-serif'
    else
      @font_family = '"Marker Felt", Helvetica, sans-serif'
    end #if
  end #def


  #
  # Locates all backups in the current user's home folder and extracts
  # notes for each one into the +output_folder+ path.  If there is more
  # than one backup present, a subfolder will be created for each individual
  # backup.
  #
  def extract_all_notes(output_folder=".")
    backups_folder = File.expand_path(
      "~/Library/Application Support/MobileSync/Backup")
    info_plists = Dir["#{backups_folder}/*/Info.plist"]
    if info_plists.size > 1
      # Create subfolders for each one based on the name and GUID found
      # in the Info.plist, and then extract each backup.
      info_plists.each do |info_plist|
        info = OSX.load_plist(File.read(info_plist))
        subdir_name = '%s (%s)' % [
          info["Device Name"], info["Unique Identifier"] ]
        extract_notes_from_backup_folder(
          File.dirname(info_plist), "#{output_folder}/#{subdir_name}")
      end #each
    elsif info_plists.size == 1
      # Use the single found folder and extract directly into output folder.
      extract_notes_from_backup_folder(
        File.dirname(info_plists[0]), output_folder)
    else
      STDERR.puts "No iPhone backup folders were found."
    end #if
  end #def

  
  #
  # Extracts the notes from the given +backup_folder+ path into the
  # +output_folder+ path.
  #
  def extract_notes_from_backup_folder(backup_folder, output_folder=".")
    
    # Locate all .mddata files in the given folder and sort by size.
    # This puts all the miscellaneous tiny files (less likely to hold notes)
    # at the end of the list so that they don't get scanned most of the time.
    backup_folder = File.expand_path(backup_folder)
    backup_files = Dir["#{backup_folder}/*.mddata"].sort do |a, b|
      File.size(b) <=> File.size(a)
    end #sort
    
    # If no .mddata files were found, this may be an old-style backup.
    if backup_files.size == 0
      extract_notes_from_legacy_backup_folder(backup_folder, output_folder)
      return
    end #if
    
    # Read the .mdinfo file (which is a standard Mac OS X property list)
    # for each backup file and see if the backup is for the Notes database.
    # If so, the notes are generated from it, and the folder scanning stops.
    backup_files.each do |backup_file|
      mdinfo_file = '%s/%s.mdinfo' % [
        backup_folder, File.basename(backup_file, ".mddata") ]
      mdinfo = OSX.load_plist(File.read(mdinfo_file))
      if mdinfo["Version"] == "2.0"
        original_path = mdinfo["Path"]
      elsif mdinfo["Version"] == "3.0"
        # Path is stored in an embedded binary plist.
        metadata = OSX.load_plist(mdinfo["Metadata"].rubyString)
        original_path = metadata["Path"]
      else
        # Unrecognized version.  Stop processing.
        STDERR.puts "Unrecognized backup version: " + mdinfo["Version"]
        break
      end #if
      if original_path == "Library/Notes/notes.db"
        database_path = "#{output_folder}/notes.db"
        FileUtils.mkdir_p output_folder
        FileUtils.cp backup_file, database_path
        extract_notes_from_database(database_path, "#{output_folder}/Notes")
        break
      end #if
    end #each

  end #def

  
  #
  # Extracts the notes from the given old-style +backup_folder+ path into
  # the +output_folder+ path.
  #
  def extract_notes_from_legacy_backup_folder(backup_folder, output_folder=".")

    # Locate all .mdbackup files in the given folder and sort by size.
    # This puts all the miscellaneous tiny files (less likely to hold notes)
    # at the end of the list so that they don't get scanned most of the time.
    backup_folder = File.expand_path(backup_folder)
    backup_files = Dir["#{backup_folder}/*.mdbackup"].sort do |a, b|
      File.size(b) <=> File.size(a)
    end #sort
    
    # Read each backup file (which a standard Mac OS X property list) and see
    # if the backup is for the Notes database.  If so, the notes are generated
    # from it, and the folder scanning stops.
    backup_files.each do |backup_file|
      dict = OSX.load_plist(File.read(backup_file))
      path = dict["Path"].to_s
      if dict["Path"] == "Library/Notes/notes.db"
        database_path = "#{output_folder}/notes.db"
        FileUtils.mkdir_p output_folder
        dict["Data"].writeToFile_atomically(database_path, false)
        extract_notes_from_database(database_path, "#{output_folder}/Notes")
        break
      end #if
    end #each

  end #def
  
  
  #
  # Extracts each note from the given SQLite +database_path+ and renders
  # it to a separate HTML file in the +output_folder+ path.  Each note
  # is named after its title in the Notes app.
  #
  def extract_notes_from_database(database_path, output_folder="./Notes")
    FileUtils.mkdir_p output_folder
    SQLite3::Database.new(database_path).execute(NOTES_QUERY) do |row|
      title = row[0]
      body = row[1]
      File.open("#{output_folder}/#{title}.html", "w") do |io|
        io.puts HTML_TEMPLATE % [ @font_family, title, body ]
      end #open
    end #execute
  end #def


  #
  # Called when the script is run from the command line.
  #
  def self.main(args)
    extractor = NoteExtractor.new
    extractor.gruber_friendly = !args.delete("--gruber-friendly").nil?
    if args.size > 0
      if args.include? "--help"
        puts "Usage: #{$0} [path/to/backup/folder]"
        puts "A single backup folder can be specified to extract one set of notes."
        puts "The backup folder can be omitted to automatically find all backups."
        puts "The --gruber-friendly option disables Marker Felt in the output."
      else
        extractor.extract_notes_from_backup_folder(args[0])
      end #if
    else
      extractor.extract_all_notes
    end #if
  end #def

end #class


# Process arguments when called from the command-line.
NoteExtractor.main(ARGV) if __FILE__ == $0
